unit ExceptionProfiler;

interface

uses
  Windows, ActiveX, Classes, ComObj, BaseProfiler, CorProf;

type
  TExceptionProfiler = class(TBaseProfiler, ICorProfilerCallback)
  protected
    // STARTUP/SHUTDOWN EVENTS
    procedure OnInitialize; override;
    procedure OnShutDown; override;
    // COR_PRF_MONITOR_EXCEPTIONS
    procedure ExceptionThrown(thrownObjectId: ULONG); safecall;
    procedure ExceptionSearchFunctionEnter(functionId: ULONG); safecall;
    procedure ExceptionSearchFunctionLeave; safecall;
    procedure ExceptionSearchFilterEnter(functionId: ULONG); safecall;
    procedure ExceptionSearchFilterLeave; safecall;
    procedure ExceptionSearchCatcherFound(functionId: ULONG); safecall;
    procedure ExceptionUnwindFunctionEnter(functionId: ULONG); safecall;
    procedure ExceptionUnwindFunctionLeave; safecall;
    procedure ExceptionUnwindFinallyEnter(functionId: ULONG); safecall;
    procedure ExceptionUnwindFinallyLeave; safecall;
    procedure ExceptionCatcherEnter(functionId: ULONG; objectId: ULONG); safecall;
    procedure ExceptionCatcherLeave; safecall;
    // COR_PRF_MONITOR_CLR_EXCEPTIONS
    procedure ExceptionCLRCatcherFound; safecall;
    procedure ExceptionCLRCatcherExecute; safecall;
  end;

const
  ProfilerCoClassName = 'ExceptionProfiler';
  ProfilerDescription = 'Exception profiler';
  ProfilerGUID: TGUID = '{0B97459B-865C-4E6B-BDBC-8EF2980A4059}';
  ProfilerClass: TComClass = TExceptionProfiler;

implementation

uses
  ComServ, SysUtils;

{ TExceptionProfiler }

procedure TExceptionProfiler.OnInitialize;
begin
  EventMask := COR_PRF_MONITOR_EXCEPTIONS or COR_PRF_MONITOR_CLR_EXCEPTIONS;
end;

procedure TExceptionProfiler.OnShutDown;
begin
  //Nothing to do
end;

procedure TExceptionProfiler.ExceptionThrown(thrownObjectId: ULONG);
var
  ClassName: WideString;
begin
  SyncEnter;
  if GetClassFromObjectId(thrownObjectId, ClassName) then
    Log('Exception thrown, name %s', [ClassName])
  else
    Log('Exception thrown, ID: $%x', [thrownObjectID]);
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionSearchFunctionEnter(functionId: ULONG);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionId, ClassName, MethodName) then
    Log('Exception handler search: entering method %s.%s', [ClassName, MethodName])
  else
    Log('Exception handler search: entering method, ID: $%x', [functionID]);
  Inc(NestingLevel);
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionSearchFunctionLeave;
begin
  SyncEnter;
  Dec(NestingLevel);
  Log('Exception handler search: exiting method');
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionSearchFilterEnter(functionId: ULONG);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionId, ClassName, MethodName) then
    Log('Exception handler search: entering user filter in method %s.%s', [ClassName, MethodName])
  else
    Log('Exception handler search: entering user filter in method, ID: $%x', [functionID]);
  Inc(NestingLevel);
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionSearchFilterLeave;
begin
  SyncEnter;
  Dec(NestingLevel);
  Log('Exception handler search: entering user filter');
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionSearchCatcherFound(functionId: ULONG);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionId, ClassName, MethodName) then
    Log('Exception handler search: found catcher in method %s.%s', [ClassName, MethodName])
  else
    Log('Exception handler search: found catcher in method, ID: $%x', [functionID]);
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionUnwindFunctionEnter(functionId: ULONG);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionId, ClassName, MethodName) then
    Log('Exception unwinding: entering method %s.%s', [ClassName, MethodName])
  else
    Log('Exception winding: entering method, ID: $%x', [functionID]);
  Inc(NestingLevel);
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionUnwindFunctionLeave;
begin
  SyncEnter;
  Dec(NestingLevel);
  Log('Exception unwinding: exiting method');
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionUnwindFinallyEnter(functionId: ULONG);
var
  ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionId, ClassName, MethodName) then
    Log('Exception unwinding: entering finally in method %s.%s', [ClassName, MethodName])
  else
    Log('Exception winding: entering finally in method, ID: $%x', [functionID]);
  Inc(NestingLevel);
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionUnwindFinallyLeave;
begin
  SyncEnter;
  Dec(NestingLevel);
  Log('Exception unwinding: leaving finally');
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionCatcherEnter(functionId,
  objectId: ULONG);
var
  ObjClassName, ClassName, MethodName: WideString;
begin
  SyncEnter;
  if GetClassAndMethodFromFunctionId(functionID, ClassName, MethodName) and
     GetClassFromObjectId(ObjectId, ObjClassName) then
    Log('Exception handler for %s entered in method %s.%s', [ObjClassName, ClassName, MethodName])
  else
    Log('Exception handler for object (ID: $%x) entered in method (ID: $%x)', [objectId, functionId]);
  Inc(NestingLevel);
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionCatcherLeave;
begin
  SyncEnter;
  Dec(NestingLevel);
  Log('Exception handler exited');
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionCLRCatcherFound;
begin
  SyncEnter;
  Log('Exception handler found in CLR');
  Inc(NestingLevel);
  SyncExit;
end;

procedure TExceptionProfiler.ExceptionCLRCatcherExecute;
begin
  SyncEnter;
  Dec(NestingLevel);
  Log('Exception handler in CLR executed');
  SyncExit;
end;

initialization
  TComObjectFactory.Create(ComServer, ProfilerClass, ProfilerGUID,
    ProfilerCoClassName, ProfilerDescription, ciMultiInstance, tmFree);
end.
